@@ -1,5 +1,5 @@
 from PythonicDISORT import subroutines
-from PythonicDISORT._loop_and_assemble_results import _loop_and_assemble_results
+from PythonicDISORT._assemble_solution_functions import _assemble_solution_functions
 import scipy as sc
 from math import pi
 from numpy.polynomial.legendre import Legendre
@@ -22,9 +22,9 @@ def pydisort(
     only_flux=False,
     f_arr=0, 
     NT_cor=False,
-    Leg_coeffs_BDRF=np.array([]),
+    BDRF_Fourier_modes=[],
     s_poly_coeffs=np.array([[]]),
-    use_sparse_NLayers=6
+    use_sparse_NLayers=13,
 ):
     """Solves the 1D RTE for the fluxes, and optionally intensity,
     of a multi-layer atmosphere with the specified optical properties, boundary conditions
@@ -47,7 +47,7 @@ def pydisort(
     Leg_coeffs_all : ndarray
         All available unweighted phase function Legendre coefficients.
     mu0 : float
-        Polar angle of the incident beam.
+        Cosine of polar angle of the incident beam.
     I0 : float
         Intensity of the incident beam.
     phi0 : float
@@ -55,7 +55,7 @@ def pydisort(
     NLeg : optional, int
         Number of phase function Legendre coefficients.
     NLoops : optional, int
-        Number of outermost loops to perform, also number of Fourier modes in the numerical solution.
+        Number of Fourier modes to use in the intensity function.
     b_pos : optional, 2darray or float
         Dirichlet boundary condition for the upward direction.
     b_neg : optional, 2darray or float
@@ -66,8 +66,9 @@ def pydisort(
         Fractional scattering into peak for each atmospheric layer.
     NT_cor : optional, bool
         Perform Nakajima-Tanaka intensity corrections?
-    Leg_coeffs_BDRF : optional, array
-        Unweighted BDRF Legendre coefficients.
+    BDRF_Fourier_modes : optional, list
+        BDRF Fourier modes, each a function with arguments mu, -mup of type array
+        and which output has the same dimensions as the outer product of the two arrays.
     s_poly_coeffs : optional, array
         Polynomial coefficients of isotropic internal sources.
         Arrange coefficients from lowest order term to highest.
@@ -79,19 +80,24 @@ def pydisort(
     array
         All mu (cosine of polar angle) quadrature nodes.
     function
-        Flux function with argument tau (type: array) for positive (upward) mu values.
+        (Energetic) Flux function with argument tau (type: array) for positive (upward) mu values.
         Returns the diffuse flux magnitudes (type: array).
     function
-        Flux function with argument tau (type: array)  for negative (downward) mu values.
+        (Energetic) Flux function with argument tau (type: array) for negative (downward) mu values.
         Returns a tuple of the diffuse and direct flux magnitudes respectively (type: (array, array)).
+    function
+        Zeroth Fourier mode of the intensity with argument tau (type: array).
+        Returns an ndarray with axes corresponding to variation with mu and tau respectively.
+        This function is useful for calculating actinic fluxes and other quantities of interest,
+        but reclassification of delta-scaled flux and other corrections must be done manually.
     function, optional
         Intensity function with arguments (tau, phi, return_Fourier_error=False) of types (array, array, bool).
-        Returns an ndarray with axes corresponding to (mu, tau, phi) variation.
+        Returns an ndarray with axes corresponding to variation with mu, tau, phi respectively.
         The optional flag `return_Fourier_error` determines whether the function will also return
         the Cauchy / Fourier convergence evaluation (type: float) for the last Fourier term.
 
     """
-    # Turn scalars to arrays
+    # Turn scalars into arrays
     # --------------------------------------------------------------------------------------------------------------------------
     tau_arr = np.atleast_1d(tau_arr)
     omega_arr = np.atleast_1d(omega_arr)
@@ -102,16 +108,27 @@ def pydisort(
     
     # Setup
     # --------------------------------------------------------------------------------------------------------------------------
-    NLayers = len(tau_arr)
     if NLeg is None:
         NLeg = NQuad
     if NLoops is None:
         NLoops = NQuad
+    if np.all(b_pos == 0):
+        b_pos = 0
+    if np.all(b_neg == 0):
+        b_neg = 0
+    if np.all(s_poly_coeffs == 0):
+        Nscoeffs = 0
+    else:
+        Nscoeffs = np.shape(s_poly_coeffs)[1]
+    NLayers = len(tau_arr)
     scalar_b_pos = False
     scalar_b_neg = False
     thickness_arr = np.concatenate([[tau_arr[0]], np.diff(tau_arr)])
-    Nscoeffs = np.shape(s_poly_coeffs)[1]
     NLeg_all = np.shape(Leg_coeffs_all)[1]
+    N = NQuad // 2
+    beam_source_bool = I0 > 0
+    iso_source_bool = Nscoeffs > 0
+    multilayer_bool = NLayers > 1
     # --------------------------------------------------------------------------------------------------------------------------
 
     # Input checks
@@ -125,15 +142,16 @@ def pydisort(
     # There must be a positive number of Legendre coefficients each with magnitude <= 1
     # The user must supply at least as many phase function Legendre coefficients as intended for use
     assert NLeg > 0
+    assert np.all(Leg_coeffs_all[:, 0] == 1)
     assert np.all(np.abs(Leg_coeffs_all) <= 1)
-    assert np.all(np.abs(Leg_coeffs_BDRF) <= 1)
+    assert np.all(np.array([f.__code__.co_argcount for f in BDRF_Fourier_modes]) == 2)
     assert NLeg <= NLeg_all
     # Ensure that the first dimension of the following inputs corresponds to the number of layers
     assert np.shape(Leg_coeffs_all)[0] == NLayers
     assert len(omega_arr) == NLayers
     if len(f_arr) != 1 or f_arr[0] != 0:
         assert len(f_arr) == NLayers
-    if Nscoeffs > 0:
+    if iso_source_bool:
         assert np.shape(s_poly_coeffs)[0] == NLayers
     # Conditions on the number of quadrature angles (NQuad), Legendre coefficients (NLeg) and loops (NLoops)
     assert NQuad >= 2
@@ -142,10 +160,9 @@ def pydisort(
     assert NLoops <= NLeg
     # Not strictly necessary but there will be tremendous inaccuracies if this is violated
     assert NQuad >= NLeg
-    N = NQuad // 2
     # We require principal angles and a downward incident beam
     assert I0 >= 0
-    if I0 > 0:
+    if beam_source_bool:
         assert 0 < mu0 and mu0 <= 1
         assert 0 <= phi0 and phi0 < 2 * pi
     # Ensure that the BC inputs are of the correct shape
@@ -165,10 +182,14 @@ def pydisort(
     
     # Some more setup
     # --------------------------------------------------------------------------------------------------------------------------
-    NBDRF = len(Leg_coeffs_BDRF)
-    weighted_Leg_coeffs_BDRF = (2 * np.arange(NBDRF) + 1) * Leg_coeffs_BDRF
+    NBDRF = len(BDRF_Fourier_modes)
     weighted_Leg_coeffs_all = (2 * np.arange(NLeg_all) + 1) * Leg_coeffs_all
     Leg_coeffs = Leg_coeffs_all[:, :NLeg]
+    if (scalar_b_pos and b_pos == 0) and (scalar_b_neg and b_neg == 0) and not iso_source_bool and beam_source_bool:
+        I0_orig = I0
+        I0 = 1
+    else:
+        I0_orig = 1
     # --------------------------------------------------------------------------------------------------------------------------
     
     # Generation of Double Gauss-Legendre quadrature weights and points
@@ -179,7 +200,7 @@ def pydisort(
     M_inv = 1 / mu_arr_pos
     
     # We do not allow mu0 to equal a quadrature / computational angle
-    assert not np.any(np.isclose(mu_arr_pos, mu0))
+    assert not np.any(np.abs(mu_arr_pos - mu0) < 1e-8)
     # --------------------------------------------------------------------------------------------------------------------------
 
     # Delta-M scaling; there is no scaling if f = 0
@@ -202,12 +223,12 @@ def pydisort(
         scaled_omega_arr = omega_arr
     # --------------------------------------------------------------------------------------------------------------------------
     
-    if NT_cor and not only_flux and I0 > 0 and np.any(f_arr > 0) and NLeg < NLeg_all:
+    if NT_cor and not only_flux and beam_source_bool and np.any(f_arr > 0) and NLeg < NLeg_all:
         
         ############################### Perform NT corrections on the intensity but not the flux ###################################
         
         # Delta-M scaled solution; no further corrections to the flux
-        flux_up, flux_down, u_star = _loop_and_assemble_results(
+        flux_up, flux_down, u0, u_star = _assemble_solution_functions(
             scaled_omega_arr,
             tau_arr,
             scaled_tau_arr_with_0,
@@ -215,16 +236,19 @@ def pydisort(
             M_inv, W,
             N, NQuad, NLeg, NLoops,
             NLayers, NBDRF,
+            multilayer_bool,
             weighted_scaled_Leg_coeffs,
-            weighted_Leg_coeffs_BDRF,
-            mu0, I0, phi0,
+            BDRF_Fourier_modes,
+            mu0, I0, I0_orig, phi0,
+            beam_source_bool,
             b_pos, b_neg,
             scalar_b_pos, scalar_b_neg,
             s_poly_coeffs,
             Nscoeffs,
+            iso_source_bool,
             scale_tau,
-            False,
-            use_sparse_NLayers
+            only_flux,
+            use_sparse_NLayers,
         )
         
         # TMS correction
@@ -272,8 +296,6 @@ def pydisort(
                 * (mu0 / (mu0 + mu_arr))[:, None, None]
                 * (p_true / (1 - f_arr[None, :, None]) - p_trun)
             )
-            mathscr_B_pos = mathscr_B[:N, :, :]
-            mathscr_B_neg = mathscr_B[N:, :, :]
             # --------------------------------------------------------------------------------------------------------------------------
             
             TMS_correction_pos = (
@@ -292,63 +314,91 @@ def pydisort(
             )
             
             # Contribution from other layers
+            # TODO: Despite being vectorized and otherwise optimized, this block of code is slower than desired
+            # and is the bottleneck for the TMS correction
             # --------------------------------------------------------------------------------------------------------------------------
-            if NLayers > 1:
+            if multilayer_bool:
 
-                def Contribution_from_layer(j):
-                    contribution = np.zeros((NQuad, Ntau, Nphi))
-                    pos_contribution_bools = l < j
-                    neg_contribution_bools = l > j
-                    scaled_tau_l_pos = scaled_tau[None, pos_contribution_bools]
-                    scaled_tau_l_neg = scaled_tau[None, neg_contribution_bools]
-                    if np.any(pos_contribution_bools):
-                        contribution[:N, pos_contribution_bools, :] = (
-                            mathscr_B_pos[:, [j], :]
-                            * (
-                                np.exp(
-                                    (scaled_tau_l_pos - scaled_tau_arr_with_0[j])
-                                    / mu_arr_pos[:, None]
-                                    - scaled_tau_arr_with_0[j] / mu0
-                                )
-                                - np.exp(
-                                    (scaled_tau_l_pos - scaled_tau_arr_with_0[j + 1])
-                                    / mu_arr_pos[:, None]
-                                    - scaled_tau_arr_with_0[j + 1] / mu0
-                                )
-                            )[:, :, None]
-                        )
-                    if np.any(neg_contribution_bools):
-                        contribution[N:, neg_contribution_bools, :] = (
-                            mathscr_B_neg[:, [j], :]
-                            * (
-                                np.exp(
-                                    (scaled_tau_arr_with_0[j + 1] - scaled_tau_l_neg)
-                                    / mu_arr_pos[:, None]
-                                    - scaled_tau_arr_with_0[j + 1] / mu0
-                                )
-                                - np.exp(
-                                    (scaled_tau_arr_with_0[j] - scaled_tau_l_neg)
-                                    / mu_arr_pos[:, None]
-                                    - scaled_tau_arr_with_0[j] / mu0
-                                )
-                            )[:, :, None]
-                        )
-                    return contribution
+                layers_arr = np.arange(NLayers)[:, None]
+                pos_contribution_mask = (l[None, :] < layers_arr).flatten()
+                neg_contribution_mask = (l[None, :] > layers_arr).flatten()
 
-                Contribution_from_other_layers = np.sum(
-                    list(map(Contribution_from_layer, range(NLayers))), axis=0
-                )
+                layers_arr_repeat = np.repeat(layers_arr, Ntau)
+                scaled_tau_tile = np.tile(scaled_tau, NLayers)
+                scaled_tau_arr_with_0_repeat = np.repeat(scaled_tau_arr_with_0, Ntau)
+
+                if np.any(pos_contribution_mask):
+                    contribution_from_other_layers_pos = np.zeros((N, NLayers * Ntau, Nphi))
+                    scaled_tau_l_tile_pos = scaled_tau_tile[pos_contribution_mask]
+                    mathscr_B_l_repeat_posmasked = mathscr_B[:N, layers_arr_repeat[pos_contribution_mask], :]
+                    scaled_tau_arr_with_0_l_tile_posmasked = scaled_tau_arr_with_0_repeat[:-Ntau][
+                        pos_contribution_mask
+                    ]
+                    scaled_tau_arr_with_0_lp1_tile_posmasked = scaled_tau_arr_with_0_repeat[Ntau:][
+                        pos_contribution_mask
+                    ]
+                    
+                    contribution_from_other_layers_pos[:, pos_contribution_mask, :] = (
+                        mathscr_B_l_repeat_posmasked
+                        * (
+                            np.exp(
+                                (scaled_tau_l_tile_pos - scaled_tau_arr_with_0_l_tile_posmasked)
+                                / mu_arr_pos[:, None]
+                                - scaled_tau_arr_with_0_l_tile_posmasked / mu0
+                            )
+                            - np.exp(
+                                (scaled_tau_l_tile_pos - scaled_tau_arr_with_0_lp1_tile_posmasked)
+                                / mu_arr_pos[:, None]
+                                - scaled_tau_arr_with_0_lp1_tile_posmasked / mu0
+                            )
+                        )[:, :, None]
+                    )
+                    contribution_from_other_layers_pos = np.sum(
+                        contribution_from_other_layers_pos.reshape((N, NLayers, Ntau, Nphi)), axis=1
+                    )
+                if np.any(neg_contribution_mask):
+                    contribution_from_other_layers_neg = np.zeros((N, NLayers * Ntau, Nphi))
+                    scaled_tau_l_tile_neg = scaled_tau_tile[neg_contribution_mask]
+                    mathscr_B_l_repeat_negmasked = mathscr_B[N:, layers_arr_repeat[neg_contribution_mask], :]
+                    scaled_tau_arr_with_0_l_tile_negmasked = scaled_tau_arr_with_0_repeat[:-Ntau][
+                        neg_contribution_mask
+                    ]
+                    scaled_tau_arr_with_0_lp1_tile_negmasked = scaled_tau_arr_with_0_repeat[Ntau:][
+                        neg_contribution_mask
+                    ]
+                    
+                    contribution_from_other_layers_neg[:, neg_contribution_mask, :] = (
+                        mathscr_B_l_repeat_negmasked
+                        * (
+                            np.exp(
+                                (scaled_tau_arr_with_0_lp1_tile_negmasked - scaled_tau_l_tile_neg)
+                                / mu_arr_pos[:, None]
+                                - scaled_tau_arr_with_0_lp1_tile_negmasked / mu0
+                            )
+                            - np.exp(
+                                (scaled_tau_arr_with_0_l_tile_negmasked - scaled_tau_l_tile_neg)
+                                / mu_arr_pos[:, None]
+                                - scaled_tau_arr_with_0_l_tile_negmasked / mu0
+                            )
+                        )[:, :, None]
+                    )
+                    contribution_from_other_layers_neg = np.sum(
+                        contribution_from_other_layers_neg.reshape((N, NLayers, Ntau, Nphi)), axis=1
+                    )
+                
                 return (
                     mathscr_B[:, l, :]
-                    * np.vstack([TMS_correction_pos, TMS_correction_neg])[:, :, None]
-                    + Contribution_from_other_layers
+                    * np.vstack((TMS_correction_pos, TMS_correction_neg))[:, :, None]
+                    + np.concatenate(
+                        (contribution_from_other_layers_pos, contribution_from_other_layers_neg), axis=0
+                    )
                 )
             # --------------------------------------------------------------------------------------------------------------------------
                 
             else:
                 return (
                     mathscr_B[:, l, :]
-                    * np.vstack([TMS_correction_pos, TMS_correction_neg])[:, :, None]
+                    * np.vstack((TMS_correction_pos, TMS_correction_neg))[:, :, None]
                 )
         # --------------------------------------------------------------------------------------------------------------------------
 
@@ -397,26 +447,28 @@ def pydisort(
             # We provide two options below, comment and uncomment as desired.
             # Option 2 is more computationally efficient but would prevent the use of autograd for testing.
 
-            NT_corrections = NT_corrections + np.concatenate(
-                [np.zeros((N, len(tau), len(phi))), IMS_correction(tau, phi)], axis=0
-            )  # Option 1
-
-            #NT_corrections[N:, :, :] += IMS_correction(tau, phi)  # Option 2
+            #NT_corrections = NT_corrections + np.concatenate(
+            #    [np.zeros((N, len(tau), len(phi))), IMS_correction(tau, phi)], axis=0
+            #)  # Option 1
 
+            NT_corrections[N:, :, :] += IMS_correction(tau, phi)  # Option 2
+               
             if return_Fourier_error:
                 u_star_outputs = u_star(tau, phi, True)
                 return (
-                    u_star_outputs[0] + np.squeeze(NT_corrections),
+                    u_star_outputs[0] + I0_orig * np.squeeze(NT_corrections),
                     u_star_outputs[1],
                 )
             else:
-                return u_star(tau, phi, False) + np.squeeze(NT_corrections)
+                return u_star(tau, phi, False) + I0_orig * np.squeeze(NT_corrections)
         # --------------------------------------------------------------------------------------------------------------------------
 
-        return mu_arr, flux_up, flux_down, u_corrected
+        return mu_arr, flux_up, flux_down, u0, u_corrected
         
     else:
-        return (mu_arr,) + _loop_and_assemble_results(
+        if only_flux:
+            NLoops = 1 # We only need to solve for the 0th Fourier mode to compute the flux
+        return (mu_arr,) + _assemble_solution_functions(
             scaled_omega_arr,
             tau_arr,
             scaled_tau_arr_with_0,
@@ -424,14 +476,17 @@ def pydisort(
             M_inv, W,
             N, NQuad, NLeg, NLoops,
             NLayers, NBDRF,
+            multilayer_bool,
             weighted_scaled_Leg_coeffs,
-            weighted_Leg_coeffs_BDRF,
-            mu0, I0, phi0,
+            BDRF_Fourier_modes,
+            mu0, I0, I0_orig, phi0,
+            beam_source_bool,
             b_pos, b_neg,
             scalar_b_pos, scalar_b_neg,
             s_poly_coeffs,
             Nscoeffs,
+            iso_source_bool,
             scale_tau,
             only_flux,
-            use_sparse_NLayers
+            use_sparse_NLayers,
         )
\ No newline at end of file
